from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.http import urlsafe_base64_encode as b64e
from enumfields.fields import EnumFieldMixin
from rest_auth.serializers import (
    LoginSerializer as RALoginSerializer,
    PasswordResetSerializer as RAPasswordResetSerializer)
from rest_framework.serializers import ModelSerializer as BaseModelSerializer

from .fields import LocalDateTimeField, RestEnumField


class ModelSerializer(BaseModelSerializer):
    serializer_field_mapping = BaseModelSerializer.serializer_field_mapping.copy()
    serializer_field_mapping.update({models.DateTimeField: LocalDateTimeField})

    def build_standard_field(self, field_name, model_field):
        cls, kwargs = super().build_standard_field(field_name, model_field)
        if isinstance(model_field, EnumFieldMixin):
            cls = RestEnumField
            kwargs['enum_type'] = model_field.enum
        return cls, kwargs


class UserDetailsSerializer(ModelSerializer):
    class Meta:
        model = get_user_model()
        read_only_fields = fields = list(set([
            'id', model.USERNAME_FIELD, model.EMAIL_FIELD]))


class LoginSerializer(RALoginSerializer):  # pylint: disable=abstract-method
    def validate_email(self, value):  # pylint: disable=no-self-use
        return value.lower()


class PasswordResetSerializer(RAPasswordResetSerializer):  # pylint: disable=abstract-method
    def get_email_context(self):
        return {
            'email_encoded': b64e(self.data['email'].lower().encode('utf-8')),
            'frontend_url': settings.DJCORE.FRONTEND_URL,
            'site_url': settings.DJCORE.SITE_URL,
        }

    def get_email_options(self):  # pylint: disable=no-self-use
        return {**super().get_email_options(), **{
            'email_template_name': 'registration/password_reset_email.txt',
            'html_email_template_name': 'registration/password_reset_email.html',
            'extra_email_context': self.get_email_context(),
        }}

    def validate_email(self, value):  # pylint: disable=no-self-use
        return super().validate_email(value.lower())



from django.core.exceptions import PermissionDenied
from django.http import Http404
from rest_framework import exceptions
from rest_framework.metadata import SimpleMetadata
from rest_framework.request import clone_request

from .fields import RestEnumField, ModelChoicesFieldMixin


class ModelChoicesMetadata(SimpleMetadata):
    """Populate choices with the values from the queryset."""
    # TODO: Make it so any number of attributes can be looked up
    def get_field_info(self, field):
        response = super().get_field_info(field)
        if isinstance(field, ModelChoicesFieldMixin):
            value_attr = field.value_attr
            str_attr = field.str_attr
            choices = [
                {value_attr: getattr(item, value_attr),
                 'repr' if str_attr is None else str_attr:
                 str(item) if str_attr is None else getattr(item, str_attr)}
                for item in field.queryset.all()
            ]
            response['choices'] = choices
        elif isinstance(field, RestEnumField):
            response['choices'] = field.choices
        return response

    def determine_actions(self, request, view):
        """Allow all allowed methods"""
        from rest_framework.generics import GenericAPIView
        actions = {}
        excluded_methods = {'HEAD', 'OPTIONS', 'PATCH', 'DELETE'}
        for method in set(view.allowed_methods) - excluded_methods:
            view.request = clone_request(request, method)
            try:
                if isinstance(view, GenericAPIView):
                    has_object = view.lookup_url_kwarg or view.lookup_field in view.kwargs
                elif method in {'PUT', 'POST'}:
                    has_object = method in {'PUT'}
                else:
                    continue
                # Test global permissions
                if hasattr(view, 'check_permissions'):
                    view.check_permissions(view.request)
                # Test object permissions
                if has_object and hasattr(view, 'get_object'):
                    view.get_object()
            except (exceptions.APIException, PermissionDenied, Http404):
                pass
            else:
                # If user has appropriate permissions for the view, include
                # appropriate metadata about the fields that should be supplied.
                serializer = view.get_serializer()
                actions[method] = self.get_serializer_info(serializer)
            finally:
                view.request = request

        return actions



from django.utils import timezone
from rest_framework.fields import ChoiceField, DateTimeField
from rest_framework.serializers import PrimaryKeyRelatedField


class LocalDateTimeField(DateTimeField):
    def default_timezone(self):
        return timezone.get_current_timezone()

    def to_representation(self, value):
        return super().to_representation(timezone.localtime(value))


class RestEnumField(ChoiceField):
    default_error_messages = ChoiceField.default_error_messages.copy()
    default_error_messages.update({'invalid': "No matching enum type."})

    def __init__(self, **kwargs):
        self.enum_type = kwargs.pop("enum_type")
        kwargs.pop("choices", None)
        super().__init__(self.enum_type.choices(), **kwargs)

    def to_internal_value(self, data):
        try:
            data_int = int(data)
        except ValueError:
            data_int = None
        for choice in self.enum_type:
            if choice.name == data or str(choice.value) == data or choice == data_int:
                return choice
        self.fail('invalid')

    def to_representation(self, value):
        return None if value is None else value.value


class ModelChoicesFieldMixin():
    def __init__(self, **kwargs):
        self.value_attr = kwargs.pop('value_attr', 'id')
        self.str_attr = kwargs.pop('str_attr', None)
        self.filtered_queryset = kwargs.pop('filtered_queryset', kwargs.get('queryset', None))
        super().__init__(**kwargs)


class ModelChoicesPrimaryKeyRelatedField(ModelChoicesFieldMixin, PrimaryKeyRelatedField):
    pass



from datetime import timedelta

from dj_core.config import Config as BaseConfig


class Config(BaseConfig):
    defaults_dict = BaseConfig.defaults_dict.copy()
    defaults_dict.simple.update({
        'DJCORE_ACCOUNT_AUTHENTICATION_METHOD': 'email',
        'DJCORE_ACCOUNT_EMAIL_REQUIRED': True,
        'DJCORE_ACCOUNT_EMAIL_VERIFICATION': 'none',
        'DJCORE_ACCOUNT_USER_MODEL_USERNAME_FIELD': None,
        'DJCORE_ACCOUNT_USERNAME_REQUIRED': False,
        'DJCORE_ACCOUNT_REGISTRATION': 'enabled',
        'DJCORE_REST_USE_JWT': True,
        'DJCORE_SWAGGER_SETTINGS': {'api_version': '1'},
        'DJCORE_JWT_AUTH': {
            'JWT_EXPIRATION_DELTA': timedelta(hours=24),
            'JWT_AUTH_HEADER_PREFIX': 'Token',
        }
    })

    def get_installed_apps(self, settings):
        return super(Config, self).get_installed_apps(settings) + [
            'dj_core_drf',
            'rest_framework',
            'rest_framework.authtoken',
            'rest_framework_swagger',
            'rest_auth',
            'allauth',
            'allauth.account',
            'rest_auth.registration',
            'revproxy',
            'django_filters',
        ]

    def get_drf_settings(self, settings):  # pylint: disable=unused-argument,no-self-use
        return {
            'DEFAULT_PERMISSION_CLASSES': [
                'rest_framework.permissions.IsAuthenticatedOrReadOnly'
            ],
            'DEFAULT_AUTHENTICATION_CLASSES': [
                'rest_framework_jwt.authentication.JSONWebTokenAuthentication',
                'rest_framework.authentication.BasicAuthentication',
                'rest_framework.authentication.SessionAuthentication',
            ],
            'DEFAULT_PAGINATION_CLASS': 'dj_core_drf.pagination.ThousandMaxLimitOffsetPagination',
            'DEFAULT_FILTER_BACKENDS': [
                'rest_framework_filters.backends.DjangoFilterBackend',
            ],
            'DEFAULT_METADATA_CLASS': 'dj_core_drf.metadata.ModelChoicesMetadata',
        }

    def get_rest_auth_serializers(self, settings):  # pylint: disable=unused-argument,no-self-use
        return {
            'USER_DETAILS_SERIALIZER': 'dj_core_drf.serializers.UserDetailsSerializer',
            'PASSWORD_RESET_SERIALIZER': 'dj_core_drf.serializers.PasswordResetSerializer',
            'LOGIN_SERIALIZER': 'dj_core_drf.serializers.LoginSerializer',
        }

    def get_settings(self):
        settings = super(Config, self).get_settings()
        settings.REST_FRAMEWORK = self.get_drf_settings(settings)
        settings.REST_AUTH_SERIALIZERS = self.get_rest_auth_serializers(settings)
        return settings
